#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""User configuration."""

import yaml


class _ConfigSingleton(type):
    """Config singleton metaclass."""

    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(_ConfigSingleton, cls).__call__(*args, **kwargs)
        return cls._instance

    @classmethod
    def set_config(mcs, config):
        """Set config which should be used in within Config singleton.

        :param config: configuration that should be used
        :type config: dict
        """
        # set _config before the singleton is instantiated
        assert mcs._instance is None  # nosec
        mcs._config = config


class Config(metaclass=_ConfigSingleton):
    """Configuration supplied by user."""

    _config = None

    def __init__(self):
        """Instantiate configuration."""
        self._raw_config = None

        if self._config is None:
            return

        with open(self._config) as input_file:
            self._raw_config = yaml.load(input_file, Loader=yaml.SafeLoader)

    def style_task(self):
        """Return style for tasks in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('task', {})
        return ret if ret is not None else {}

    def style_flow(self):
        """Return style for a flow node in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        default = {'style': 'filled', 'fillcolor': '#CCCCCC'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default
        ret = self._raw_config['style'].get('flow', default)
        return ret if ret is not None else default

    def style_condition(self):
        """Return style for conditions in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        default_style = {'shape': 'octagon'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('condition', default_style)
        return ret if ret is not None else default_style

    def style_condition_foreach(self):
        """Return style for foreach edges in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        default_style = {'shape': 'doubleoctagon'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('condition', default_style)
        return ret if ret is not None else default_style

    def style_storage(self):
        """Return style for storage in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        default_style = {'shape': 'folder', 'color': '#AAAAAA'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('storage', default_style)
        return ret if ret is not None else {}

    def style_edge(self):
        """Return style for edges in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('edge', {})
        return ret if ret is not None else {}

    def style_store_edge(self):
        """Return style for edges that lead to a storage in the graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#AAAAAA', 'style': 'dashed'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('store-edge', default_style)
        return ret if ret is not None else default_style

    def style_graph(self):
        """Return style for the whole graph, see graphviz styling options.

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('graph', {})
        return ret if ret is not None else {}

    def style_fallback_edge(self):
        """Return style for fallback edges.

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#FF0000', 'style': 'dashed'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style

    def style_fallback_node(self):
        """Return style for fallback node.

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#FF0000', 'shape': 'point'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style

    def style_fallback_true(self):
        """Return style for fallback true node.

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#222222', 'shape': 'doubleoctagon'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""User's global configuration section parsed from YAML config file."""

from .errors import ConfigurationError
from .helpers import check_conf_keys


class GlobalConfig(object):
    """User global configuration stated in YAML file."""

    DEFAULT_CELERY_QUEUE = 'celery'
    predicates_module = 'selinonlib.predicates'

    default_task_queue = DEFAULT_CELERY_QUEUE
    default_dispatcher_queue = DEFAULT_CELERY_QUEUE
    migration_dir = None

    _trace_logging = []
    _trace_function = []
    _trace_storage = []
    _trace_sentry = []

    def __init__(self):
        """Placeholder."""
        raise NotImplementedError("Cannot instantiate global config")

    @classmethod
    def dump_trace(cls, output, config_name, indent_count=0):
        """Dump trace configuration to output stream.

        :param output: output stream to write to
        :param config_name: name of configuration class instance to be referenced when initializing trace
        :param indent_count: indentation that should be used to indent source
        """
        indent = indent_count * 4 * " "

        for _ in cls._trace_logging:
            output.write('%s%s.trace_by_logging()\n' % (indent, config_name))

        for entry in cls._trace_storage:
            output.write('%s%s.trace_by_func(functools.partial(%s.%s, %s))\n'
                         % (indent, config_name, entry[0].class_name, entry[1], entry[0].var_name))

        for entry in cls._trace_function:
            output.write('%sfrom %s import %s\n' % (indent, entry[0], entry[1]))
            output.write('%s%s.trace_by_func(%s)\n' % (indent, config_name, entry[1]))

        for entry in cls._trace_sentry:
            output.write("%s%s.trace_by_sentry(dsn=%s)\n"
                         % (indent, config_name, "'%s'" % entry if entry is not True else None))

    @classmethod
    def _parse_trace_storage(cls, trace_def, system):
        """Parse tracing by storage.

        :param trace_def: definition of tracing as supplied in the YAML file
        :param system: system instance
        """
        if not isinstance(trace_def, dict):
            raise ConfigurationError("Configuration of storage trace expects dict, got '%s' instead (type: %s)"
                                     % (trace_def, type(trace_def)))

        if 'name' not in trace_def:
            raise ConfigurationError('Expected storage name in tracing configuration, got %s instead'
                                     % trace_def)

        unknown_conf = check_conf_keys(trace_def, known_conf_opts=('method', 'name'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration for trace storage '%s' supplied: %s"
                                     % (trace_def, unknown_conf))

        cls._trace_storage.append((system.storage_by_name(trace_def['name']), trace_def.get('method', 'trace')))

    @classmethod
    def _parse_trace_function(cls, trace_def):
        """Parse tracing by external function.

        :param trace_def: definition of tracing as supplied in the YAML file
        """
        if not isinstance(trace_def, dict):
            raise ConfigurationError("Configuration of trace function expects dict, got '%s' instead (type: %s)"
                                     % (trace_def, type(trace_def)))

        if 'import' not in trace_def:
            raise ConfigurationError('Expected import definition in function trace configuration, got %s instead'
                                     % trace_def)

        if 'name' not in trace_def:
            raise ConfigurationError('Expected function name in function trace configuration, got %s instead'
                                     % trace_def)

        unknown_conf = check_conf_keys(trace_def, known_conf_opts=('import', 'name'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration for trace function '%s' from '%s' supplied: %s"
                                     % (trace_def['name'], trace_def['import'], unknown_conf))

        cls._trace_function.append((trace_def['import'], trace_def['name']))

    @classmethod
    def _parse_trace_logging(cls, trace_def):
        """Parse tracing by Python's logging facilities.

        :param trace_def: definition of tracing as supplied in the YAML file
        """
        if trace_def is True:
            cls._trace_logging.append(trace_def)

    @classmethod
    def _parse_trace_sentry(cls, trace_def):
        """Parse tracing by Sentry - error tracking software.

        :param trace_def: definition of tracing as supplied in the YAML file
        """
        if not isinstance(trace_def, dict):
            raise ConfigurationError("Configuration of Sentry tracing expects dict, got '%s' instead (type: %s)"
                                     % (trace_def, type(trace_def)))

        unknown_conf = check_conf_keys(trace_def, known_conf_opts=('dsn',))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration for Sentry trace function supplied: %s" % unknown_conf)

        cls._trace_sentry.append(trace_def.get('dsn', True))

    @classmethod
    def _parse_trace(cls, system, trace_record):
        """Parse trace configuration entry.

        :param system: system instance for which the parsing is done (for storage lookup)
        :param trace_record: trace record to be parsed
        """
        if trace_record is None:
            raise ConfigurationError('Trace not defined properly in global configuration section, '
                                     'see documentation for more info')

        if trace_record is False:
            return

        trace_record = [trace_record] if not isinstance(trace_record, list) else trace_record

        for entry in trace_record:
            if 'logging' in entry:
                cls._parse_trace_logging(entry['logging'])

            if 'storage' in entry:
                cls._parse_trace_storage(entry['storage'], system)

            if 'function' in entry:
                cls._parse_trace_function(entry['function'])

            if 'sentry' in entry:
                cls._parse_trace_sentry(entry['sentry'])

    @classmethod
    def from_dict(cls, system, dict_):
        """Parse global configuration from a dictionary.

        :param system: system instance for storage lookup
        :param dict_: dictionary containing global configuration as stated in YAML config file
        """
        if 'predicates_module' in dict_:
            cls.predicates_module = dict_.pop('predicates_module')

        if 'trace' in dict_:
            cls._parse_trace(system, dict_.pop('trace'))

        cls.default_dispatcher_queue = dict_.pop('default_dispatcher_queue', cls.DEFAULT_CELERY_QUEUE)
        cls.default_task_queue = dict_.pop('default_task_queue', cls.DEFAULT_CELERY_QUEUE)
        cls.migration_dir = dict_.pop('migration_dir', None)

        if dict_:
            raise ConfigurationError("Unknown configuration options supplied in global configuration section: %s"
                                     % dict_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Function that is run on selective flow/task run."""

from .errors import ConfigurationError
from .helpers import check_conf_keys


class SelectiveRunFunction(object):
    """Function that is run on selective flow/task run to ensure whether task/flow should be run."""

    _DEFAULT_IMPORT_PATH = 'selinonlib.routines'
    _DEFAULT_FUNCTION_NAME = 'always_run'

    def __init__(self, name, import_path):
        """Initialize selective function as stated in YAML configuration file.

        :param name: name of the selective run function
        :param import_path: import describing module from which the selective function should be imported
        """
        self.name = name
        self.import_path = import_path

    @staticmethod
    def construct_import_name(name, import_path):
        """Construct import name that will be used in generated config.

        :param name: name of the function that will be imported
        :param import_path: import that should be used to import function
        :return: string representation of function that will be used in generated config
        """
        return "_{import_path}_{name}".format(import_path=import_path.replace(".", "_"), name=name)

    def get_import_name(self):
        """Get import name that will be used in generated config.

        :return: string representation of function that will be used in generated config
        """
        return self.construct_import_name(self.name, self.import_path)

    @classmethod
    def get_default(cls):
        """Get default selective run function.

        :return: default instance that will be used if user did not provided configuration
        """
        return cls(cls._DEFAULT_FUNCTION_NAME, cls._DEFAULT_IMPORT_PATH)

    @classmethod
    def from_dict(cls, dict_):
        """Instantiate selective run function based on definition in a dictionary.

        :return: selective run function that will be used based on configuration
        :rtype: SelectiveRunFunction
        """
        if not dict_:
            return cls.get_default()

        unknown_conf = check_conf_keys(dict_, known_conf_opts=('name', 'import'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration options for selective run function supplied: %s"
                                     % (unknown_conf.keys()))

        name = dict_.pop('name', cls._DEFAULT_FUNCTION_NAME)
        import_path = dict_.pop('import', cls._DEFAULT_IMPORT_PATH)

        return cls(name, import_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Errors and exceptions that can occur in Selinonlib code base."""


class FatalTaskError(Exception):
    """An exception that is raised by task on fatal error - task will be not retried."""


class UnknownFlowError(Exception):
    """Raised if there was requested or referenced flow that is not stated in the YAML configuration file."""


class UnknownStorageError(Exception):
    """Raised if there was requested or referenced storage that is not stated in the YAML configuration file."""


class ConfigurationError(Exception):
    """Raised on errors that indicate errors in the configuration files."""


class SelectiveNoPathError(Exception):
    """Raised when there is no path in the flow to requested node in selective task runs."""


class NoParentNodeError(Exception):
    """An exception raised when requested parent node (task/flow), but no such parent defined."""


class RequestError(Exception):
    """An error raised if there was an issue with request issued by user - usually means bad usage error."""


class UnknownError(Exception):
    """An error raised on unknown scenarios - possibly some bug in code."""


class Retry(Exception):
    """Retry task as would Celery do except you can only specify countdown for retry."""

    def __init__(self, countdown):
        """Init retry.

        :param countdown: countdown in seconds
        """
        self.countdown = countdown
        Exception.__init__(self, countdown)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""A Python class abstraction."""


class TaskClass(object):
    """Task's class abstraction."""

    def __init__(self, class_name, import_path):
        """Instantiate task class representation.

        :param class_name: name of the class
        :param import_path: import to be used to import class
        """
        self.class_name = class_name
        self.import_path = import_path
        self.tasks = []

    def task_of_class(self, task):
        """Check that a task is of this class type.

        :param task: task to be checked
        :return: True if task is of this class
        """
        return self.class_name == task.class_name and self.import_path == task.import_path

    def add_task(self, task):
        """Register a task to this class.

        :param task: task to be added
        """
        assert self.task_of_class(task)  # nosec
        self.tasks.append(task)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinon SQL Database adapter - PostgreSQL."""

from selinon import DataStorage

try:
    from sqlalchemy import (create_engine, Column, Integer, Sequence, String)
    from sqlalchemy.dialects.postgresql import JSONB
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import sessionmaker
except ImportError:
    raise ImportError("Please install SQLAlchemy using `pip3 install SQLAlchemy` in order to use SQLStorage")
try:
    from sqlalchemy_utils import create_database, database_exists
except ImportError:
    raise ImportError("Please install SQLAlchemy-Utils using `pip3 install SQLAlchemy-Utils in order to use SQLStorage")

_Base = declarative_base()  # pylint: disable=invalid-name


class Result(_Base):
    """Record for a task result."""

    __tablename__ = 'result'

    id = Column(Integer, Sequence('result_id'), primary_key=True)  # pylint: disable=invalid-name
    flow_name = Column(String(128))
    task_name = Column(String(128))
    task_id = Column(String(255), unique=True)
    # We are using JSONB for postgres, if you want to use other database, change column type
    result = Column(JSONB)
    node_args = Column(JSONB)

    def __init__(self, node_args, flow_name, task_name, task_id, result):  # noqa
        self.flow_name = flow_name
        self.task_name = task_name
        self.task_id = task_id
        self.result = result
        self.node_args = node_args


class SqlStorage(DataStorage):
    """Selinon SQL Database adapter - PostgreSQL."""

    def __init__(self, connection_string, encoding='utf-8', echo=False):
        """Initialize PostgreSQL adapter from YAML configuration file.

        :param connection_string:
        :param encoding:
        :param echo:
        """
        super(SqlStorage, self).__init__()

        self.engine = create_engine(connection_string, encoding=encoding, echo=echo)
        self.session = None

    def is_connected(self):  # noqa
        return self.session is not None

    def connect(self):  # noqa
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        self.session = sessionmaker(bind=self.engine)()
        _Base.metadata.create_all(self.engine)

    def disconnect(self):  # noqa
        if self.is_connected():
            self.session.close()
            self.session = None

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        assert self.is_connected()  # nosec

        record = self.session.query(Result).filter_by(task_id=task_id).one()

        assert record.task_name == task_name  # nosec
        return record.result

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert self.is_connected()  # nosec

        record = Result(node_args, flow_name, task_name, task_id, result)
        try:
            self.session.add(record)
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

        return record.id

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Selinonlib library helpers."""

from contextlib import contextmanager
import json
from json import JSONEncoder
import logging
import os
import subprocess
import tempfile

from .errors import RequestError

_logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class _SelinonJSONEncoder(JSONEncoder):
    """Enhanced JSON serializer."""


def dict2strkwargs(dict_):
    """Convert a dictionary into arguments to a string representation that can be used as arguments to a function."""
    ret = ""
    for key, value in dict_.items():
        if ret:
            ret += ", "
        ret += "%s=%s" % (key, expr2str(value))
    return ret


def expr2str(expr):
    """Convert a Python expression into a Python code."""
    if isinstance(expr, dict):
        return str(expr)
    elif isinstance(expr, list):
        # s/'['foo']['bar']'/['foo']['bar']/ (get rid of leading ')
        return "%s" % expr
    elif isinstance(expr, str):
        return "'%s'" % expr

    # some build in type such as bool/int/...
    return "%s" % str(expr)


def keylist2str(keylist):
    """Convert keylist to a string representation.

    :param keylist: keylist to be converted
    :type keylist: list
    :return: string representation
    :rtype: str
    """
    return "".join(map(lambda x: "['" + str(x) + "']", keylist))


@contextmanager
def pushd(new_dir):
    """Traverse directory tree in push/pop manner.

    :param new_dir: new directory to cd to
    :type new_dir: str
    """
    prev_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(prev_dir)


def dict2json(dict_, pretty=True):
    """Convert dict to json (string).

    :param dict_: dictionary to be converted
    :param pretty: if True, nice formatting will be used
    :type pretty: bool
    :return: formatted dict in json
    :rtype: str
    """
    def default(obj):
        """Serialize some additional types that the default encoder does not support."""
        if isinstance(obj, set):
            return list(obj)
        raise TypeError("Cannot serialize %r of type %r" % (obj, type(obj)))

    if pretty is True:
        return json.dumps(dict_, sort_keys=True, separators=(',', ': '), indent=2, default=default)

    return json.dumps(dict_)


def get_function_arguments(func):
    """Get arguments of function.

    :param func: function to parse arguments
    :return: list of arguments that predicate function expects
    """
    return list(func.__code__.co_varnames[:func.__code__.co_argcount])


def check_conf_keys(dict_, known_conf_opts):
    """Check supplied configuration options against known configuration options.

    :param dict_: dict with configuration options
    :param known_conf_opts: known configuration options
    :return: configuration options that are now known with their values
    """
    return {k: v for k, v in dict_.items() if k not in known_conf_opts}


def git_previous_version(file_path):
    """Get hash of previous version of a file in Git CVS.

    :param file_path: a path to file to get previous version from
    :return: git hash of previous version and it's depth from master
    :rtype: tuple
    """
    cmd = "git log --max-count 1 --pretty=format:%H".split(' ')
    cmd.append(file_path)
    try:
        git_hash = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
    except subprocess.CalledProcessError as exc:
        err_msg = "Failed to get previous version of file %r using git: %s" % (file_path, str(exc.output))
        raise RequestError(err_msg) from exc

    cmd = "git rev-list --ancestry-path {}..master".format(git_hash).split(' ')
    try:
        depth = len(subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True))
    except subprocess.CalledProcessError as exc:
        err_msg = "Failed to get git log depth for file %r using git: %s" % (file_path, str(exc.output))
        raise RequestError(err_msg) from exc

    _logger.debug("Previous version found in Git CVS of file %r is %s with a depth of %d",
                  file_path, git_hash, depth)
    return git_hash, depth


def git_previous_version_file(git_hash, file_path, tmp_dir=None):
    """Get previous version of a file in Git CVS.

    :param git_hash: git hash for the given file
    :param file_path: a path to file to get previous version from
    :param tmp_dir: a directory to store the content of previous file version
    :return: a path to a temporary file with content from previous version
    """
    cmd = ["git", "show", "%s:%s" % (git_hash, file_path if file_path.startswith('/') else './' + file_path)]
    try:
        file_content = subprocess.check_output(cmd, stderr=subprocess.PIPE, universal_newlines=True)
    except subprocess.CalledProcessError as exc:
        err_msg = "Failed to get content of file %r in version %s using git: %s" \
                  % (file_path, git_hash, str(exc.output))
        raise RequestError(err_msg) from exc

    _logger.debug("Using git version %r for file %r", git_hash, file_path)
    with tempfile.NamedTemporaryFile(mode="w", dir=tmp_dir, delete=False) as temp_file:
        temp_file.write(file_content)
        return temp_file.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Most-Recently-Used cache implementation."""

from .lru import LRU


class MRU(LRU):
    """Most-Recently-Used - implementation based on LRU."""

    def _clean_cache(self):
        """Trim cache."""
        while self.current_cache_size + 1 > self.max_cache_size and self.current_cache_size > 0:
            self._remove_record(self._record_tail)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Strategies used when migrations are preformed."""

from enum import Enum

from selinon.errors import UnknownError


class TaintedFlowStrategy(Enum):
    """Strategies used when migrations are preformed."""

    # Strategies need to be numbered by priority
    IGNORE = 1
    RETRY = 2
    FAIL = 3

    @classmethod
    def get_options(cls):
        """Get all options."""
        return cls.IGNORE, cls.RETRY, cls.FAIL

    @classmethod
    def get_option_names(cls):
        """Get names of all options."""
        return list(map(lambda x: x.name, cls.get_options()))

    @classmethod
    def get_default_option(cls):
        """Get default option for migration strategies."""
        return cls.IGNORE.name

    @classmethod
    def get_option_by_name(cls, name):
        """Get option by its string representation.

        :param name: option string representation
        :return: option for the given name representation
        """
        for option in (cls.IGNORE, cls.RETRY, cls.FAIL):
            if option.name.lower() == name.lower():
                return option
        raise UnknownError("Unknown tainted flow strategy %r" % name)

    @classmethod
    def get_preferred_strategy(cls, strategy1, strategy2):
        """Get preferred strategy in case of multiple migrations being done at once.

        Chose strategy with higher priority - if two migrations are done at the same time one marked as IGNORE another
        as RETRY, we have to retry flow.

        :param strategy1: the first strategy to compare with
        :type strategy1: TaintedFlowStrategy
        :param strategy2: the second strategy to compare with
        :type strategy2: TaintedFlowStrategy
        :return: strategy with higher priority
        :rtype: TaintedFlowStrategy
        """
        return max(strategy1, strategy2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Leaf predicate in condition - should always return True/False for the given input."""

import ast
import importlib
import logging

from .errors import ConfigurationError
from .globalConfig import GlobalConfig
from .helpers import dict2strkwargs
from .helpers import get_function_arguments
from .predicate import Predicate


class LeafPredicate(Predicate):
    """Leaf predicate representation."""

    _logger = logging.getLogger(__name__)

    def __init__(self, predicate_func, node, flow, args=None):
        """Instantiate leaf predicate.

        :param predicate_func: predicate function
        :param node: node that predicate conforms to
        :param flow: flow to which predicate belongs to
        :param args: predicate arguments that should be used
        """
        super().__init__()
        self.node = node
        self.flow = flow

        self._func = predicate_func
        self._args = args or {}
        self._func_args = get_function_arguments(self._func)

        if self.requires_message() and not self.node:
            raise ConfigurationError("Expected task name for predicate '%s'" % self._func.__name__)

    def requires_message(self):
        """Check whether this predicate requires results from parent node.

        :return: True if predicate requires a message from a parent node
        """
        return 'message' in self._func_args

    def requires_node_args(self):
        """Check whether this predicate inspects arguments passed to flow.

        :return: True if predicate requires a node arguments
        """
        return 'node_args' in self._func_args

    def _check_parameters(self):
        """Check user defined predicate parameters against predicate parameters.

        :raises: ConfigurationError
        """
        func_args = get_function_arguments(self._func)
        user_args = self._args.keys()

        if 'message' in func_args:
            # message argument is implicit and does not need to be specified by user
            func_args.remove('message')

        if 'node_args' in func_args:
            # node_args are implicit as well
            func_args.remove('node_args')

        func_args = set(func_args)
        user_args = set(user_args)

        error = False

        for arg in func_args - user_args:
            self._logger.error("Argument '%s' of predicate '%s' not specified in flow '%s'",
                               arg, self._func.__name__, self.flow.name)
            error = True

        for arg in user_args - func_args:
            self._logger.error("Invalid argument '%s' for predicate '%s' in flow '%s'",
                               arg, self._func.__name__, self.flow.name)
            error = True

        if error:
            raise ConfigurationError("Bad predicate arguments specified in flow '%s'" % self.flow.name)

    def _check_usage(self):
        """Check correct predicate usage.

        :raises: ValueError
        """
        if self.requires_message() and self.node and self.node.is_flow():
            raise ConfigurationError("Results of sub-flows cannot be used in predicates")
        if self.requires_message() and not self.node:
            raise ConfigurationError("Cannot inspect results in starting edge in predicate '%s'" % self._func.__name__)
        if self.requires_message() and not self.node.storage:
            raise ConfigurationError("Cannot use predicate '%s' that requires a results "
                                     "of node '%s' (import: %s) since this node has no storage assigned"
                                     % (str(self), self.node.name, self.node.import_path))

    def check(self):
        """Check whether predicate is correctly used.

        :raises: ValueError
        """
        self._check_usage()
        self._check_parameters()

    def __str__(self):
        """Create a string representation of this predicate (Python function call).

        :return: a string representation of this predicate
        """
        if self.requires_message():
            if self._args:
                return "%s(db.get('%s'), %s)"\
                       % (self._func.__name__, self._task_str_name(), dict2strkwargs(self._args))

            return "%s(db.get('%s'))" % (self._func.__name__, self._task_str_name())

        # we hide node_args parameter
        return "%s(%s)" % (self._func.__name__, dict2strkwargs(self._args))

    def _task_str_name(self):  # noqa
        # task_name can be None if we have starting edge
        if self.node is None:
            return 'None'

        return "%s" % self.node.name

    def ast(self):
        """Create Python AST of this predicate.

        :return: AST representation of predicate
        """
        # we could directly use db[task] in predicates, but predicates should not handle database errors,
        # so leave them on higher level (selinon) and index database before predicate is being called

        kwargs = []
        # we want to avoid querying to database if possible, if a predicate does not require message, do not ask for it
        if self.requires_message():
            # this can raise an exception if check was not run, since we are accessing storage that can be None
            kwargs.append(ast.keyword(arg='message',
                                      value=ast.Call(func=ast.Attribute(value=ast.Name(id='db', ctx=ast.Load()),
                                                                        attr='get', ctx=ast.Load()),
                                                     args=[ast.Str(s=self._task_str_name())],
                                                     keywords=[], starargs=None, kwargs=None)))
        if self.requires_node_args():
            kwargs.append(ast.keyword(arg='node_args', value=ast.Name(id='node_args', ctx=ast.Load())))

        kwargs.extend([ast.keyword(arg=k, value=ast.Str(s=v)) for k, v in self._args.items()])

        return ast.Call(func=ast.Name(id=self._func.__name__, ctx=ast.Load()),
                        args=[], starargs=None, kwargs=None, keywords=kwargs)

    def predicates_used(self):
        """Return a list of predicates that are used.

        :return: list of predicates that are used
        :rtype: List[Predicate]
        """
        return [self._func] if self._func else []

    def nodes_used(self):
        """Return a list of nodes that are used by this predicate.

        :return: list of nodes that are used
        :rtype: List[Node]
        """
        return [self.node] if self.node else []

    @classmethod
    def create(cls, name, node, flow, args=None):  # pylint: disable=arguments-differ
        """Create predicate.

        :param name: predicate name
        :type name: str
        :param node: node to which predicate belongs
        :type node: Node
        :param flow: flow to which predicate belongs
        :type flow: Flow
        :param args: predicate arguments
        :return: an instantiated predicate
        :raises: ImportError
        """
        try:
            module = importlib.import_module(GlobalConfig.predicates_module)
            predicate = getattr(module, name)
        except ImportError:
            cls._logger.error("Cannot import predicate '%s'", name)
            raise
        return LeafPredicate(predicate, node, flow, args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from functools import partial
from selinonlib.strategies import (
    linear_increase,
    linear_adapt,
    biexponential_increase,
    biexponential_decrease,
    biexponential_adapt,
    random,
    constant
)
from selinonlibTestCase import SelinonlibTestCase

# Values used in tests for strategies functions
_TEST_START_RETRY = 2
_TEST_MAX_RETRY = 20
_TEST_STEP_RETRY = 5
_TEST_PREV_RETRY = 42

# Available strategies that take 1 additional arguments
_STRATEGIES_ONE_ARG = [
    partial(constant, retry=_TEST_START_RETRY)
]

# Available strategies that take 2 additional arguments
_STRATEGIES_TWO_ARG = [
    partial(biexponential_increase, start_retry=_TEST_START_RETRY, max_retry=_TEST_MAX_RETRY),
    partial(biexponential_decrease, start_retry=_TEST_START_RETRY, stop_retry=_TEST_MAX_RETRY),
    partial(biexponential_adapt, start_retry=_TEST_START_RETRY, max_retry=_TEST_MAX_RETRY),
    partial(random, start_retry=_TEST_START_RETRY, max_retry=_TEST_MAX_RETRY)
]


# Available strategies that take 3 additional arguments
_STRATEGIES_THREE_ARG = [
    partial(linear_increase, start_retry=_TEST_START_RETRY, max_retry=_TEST_MAX_RETRY, step=_TEST_STEP_RETRY),
    partial(linear_adapt, start_retry=_TEST_START_RETRY, max_retry=_TEST_MAX_RETRY, step=_TEST_STEP_RETRY)
]

_STRATEGIES_ALL_ARG = _STRATEGIES_ONE_ARG + _STRATEGIES_TWO_ARG + _STRATEGIES_THREE_ARG


@pytest.mark.parametrize("strategy", _STRATEGIES_ALL_ARG)
class TestSamplingStrategies(SelinonlibTestCase):
    def test_start(self, strategy):
        status = {
            'previous_retry': None,
            'active_nodes': ['Task1', 'Task2'],
            'failed_nodes': [],
            'new_started_nodes': ['Task1', 'Task2'],
            'new_fallback_nodes': [],
            'finished_nodes': []
        }

        retry = strategy(status)
        assert retry is not None and retry > 0

    def test_only_active(self, strategy):
        status = {
            'previous_retry': _TEST_PREV_RETRY,
            'active_nodes': ['Task1', 'Task2'],
            'failed_nodes': [],
            'new_started_nodes': [],
            'new_fallback_nodes': [],
            'finished_nodes': []
        }

        retry = strategy(status)
        assert retry is not None and retry > 0

    def test_active_and_started(self, strategy):
        status = {
            'previous_retry': _TEST_PREV_RETRY,
            'active_nodes': ['Task1', 'Task2'],
            'failed_nodes': [],
            'new_started_nodes': ['Task3'],
            'new_fallback_nodes': [],
            'finished_nodes': []
        }

        retry = strategy(status)
        assert retry is not None and retry > 0

    def test_finish(self, strategy):
        status = {
            'previous_retry': _TEST_PREV_RETRY,
            'active_nodes': [],
            'failed_nodes': [],
            'new_started_nodes': [],
            'new_fallback_nodes': [],
            'finished_nodes': []
        }

        retry = strategy(status)
        assert retry is None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Strategy for scheduling dispatcher - system state sampling."""

import importlib

from .errors import ConfigurationError
from .helpers import check_conf_keys
from .helpers import get_function_arguments


class Strategy(object):
    """Sampling strategy for Dispatcher."""

    _DEFAULT_MODULE = 'selinonlib.strategies'
    _DEFAULT_FUNCTION = 'biexponential_increase'
    _DEFAULT_FUNC_ARGS = {'start_retry': 2, 'max_retry': 120}

    _EXPECTED_STRATEGY_FUNC_ARGS = {'status'}

    def __init__(self, module=None, function=None, func_args=None):
        """Instantiate strategy function representation.

        :param module: module from which sampling strategy should be imported
        :param function: sampling function name
        :param func_args: sampling function arguments
        """
        self.module = module or self._DEFAULT_MODULE
        self.function = function or self._DEFAULT_FUNCTION
        self.func_args = func_args or self._DEFAULT_FUNC_ARGS

    @classmethod
    def from_dict(cls, strategy_dict, flow_name):
        """Parse strategy entry.

        :param strategy_dict: strategy entry in config to be parsed
        :param flow_name: name of the flow for which the strategy function should be used.
        """
        if not strategy_dict:
            return cls()

        if not isinstance(strategy_dict, dict):
            raise ConfigurationError("Strategy not defined properly in global configuration section, expected dict,"
                                     "got %s in flow %r" % (strategy_dict, flow_name))

        if 'name' not in strategy_dict:
            raise ConfigurationError("Sampling strategy stated in global configuration but no strategy"
                                     "name defined in flow %s" % flow_name)

        if not isinstance(strategy_dict['args'], dict):
            raise ConfigurationError('Arguments to strategy function should be stated as dict,'
                                     'got %s instead in flow %r' % (strategy_dict['args'], flow_name))

        strategy_module = strategy_dict.get('import', cls._DEFAULT_MODULE)

        raw_module = importlib.import_module(strategy_module)
        raw_func = getattr(raw_module, strategy_dict['name'])

        # perform checks on args supplied
        user_args_keys = strategy_dict['args'].keys()
        func_args = set(get_function_arguments(raw_func))

        if (func_args - user_args_keys) != cls._EXPECTED_STRATEGY_FUNC_ARGS:
            raise ConfigurationError('Unknown or invalid arguments supplied to sampling strategy function,'
                                     'expected %s, got %s for strategy %r in flow %r'
                                     % ((func_args - cls._EXPECTED_STRATEGY_FUNC_ARGS), set(user_args_keys),
                                        strategy_dict['name'], flow_name))

        unknown_conf = check_conf_keys(strategy_dict, known_conf_opts=('name', 'import', 'args'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration for sampling strategy %r supplied: '%s'"
                                     % (strategy_dict['name'], unknown_conf))

        return cls(strategy_module, strategy_dict['name'], strategy_dict['args'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""In memory storage implementation."""

import json as jsonlib
import sys

from selinon import DataStorage


class InMemoryStorage(DataStorage):
    """Storage that stores results in memory without persistence."""

    def __init__(self, echo=False, json=False):
        """Initialize storage, values passed from YAML cofig file.

        :param echo: echo results to stdout/stderr - provide 'stderr' or 'stdout' to echo data retrieval and storing
        :param json: if True a JSON will be printed
        """
        super(InMemoryStorage, self).__init__()
        self.database = {}
        self.echo_file = None

        if not echo and json:
            raise ValueError("JSON parameter requires echo to be specified ('stdout' or 'stderr')")

        self.echo_json = json

        if echo == 'stdout' or echo is True:
            self.echo_file = sys.stdout
        elif echo == 'stderr':
            self.echo_file = sys.stderr

    def is_connected(self):  # noqa
        return True

    def connect(self):  # noqa
        pass

    def disconnect(self):  # noqa
        pass

    def retrieve(self, flow_name, task_name, task_id):  # noqa
        try:
            result = self.database[task_id]['result']
            if self.echo_file and self.echo_json:
                jsonlib.dump(result, self.echo_file, sort_keys=True, separators=(',', ': '), indent=2)
            elif self.echo_file:
                print(result, file=self.echo_file)
        except KeyError:
            raise FileNotFoundError("Record not found in database")

    def store(self, node_args, flow_name, task_name, task_id, result):  # noqa
        assert task_id not in self.database  # nosec

        record = {
            'node_args': node_args,
            'task_name': task_name,
            'flow_name': flow_name,
            'task_id': task_id,
            'result': result
        }
        self.database[task_id] = record

        if self.echo_file and self.echo_json:
            jsonlib.dump(record, self.echo_file, sort_keys=True, separators=(',', ': '), indent=2)
        elif self.echo_file:
            print(record, file=self.echo_file)

        # task_id is unique for the record
        return task_id

    def store_error(self, node_args, flow_name, task_name, task_id, exc_info):  # noqa
        # just to make pylint happy
        raise NotImplementedError()



#!/usr/bin/env python3
"""Supportive and handling library for Selinon."""

from .builtinPredicate import AndPredicate
from .builtinPredicate import BuiltinPredicate
from .builtinPredicate import NaryPredicate
from .builtinPredicate import NotPredicate
from .builtinPredicate import OrPredicate
from .builtinPredicate import UnaryPredicate
from .codename import selinonlib_version_codename
from .config import Config
from .edge import Edge
from .errors import *  # pylint: disable=wildcard-import
from .flow import Flow
from .leafPredicate import LeafPredicate
from .node import Node
from .predicate import Predicate
from .storage import Storage
from .system import System
from .task import Task
from .version import selinonlib_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Simulate execution in a single CLI run.

Using Selinon simulator is a good way to test you configuration and behaviour locally if you would like to save
some time when debugging or exploring all the possibilities that Selinon offers you. Keep in mind that running
Selinon locally was designed for development purposes and behaviour in general can (and in many cases will) vary.

The key idea behind simulator is to simulate exchange of messages which is done by your Celery broker and Kombu under
the hood of Celery. Thus there are lazily created queues that are referenced by their names (see
selinonlib.simulator.timeQueue for implementation details). These queues hold messages prior to time on which they were
scheduled. Under the hood there is used a heap queue to optimize inserting to O(log(N)) in the worst case where N is
number messages currently in the queue. These queues are coupled into QueuePool (selinonlib.simulate.queuePool) which
encapsulates all queues, keeps their references, instantiates it lazily and provides concurrency safety.

In order to avoid starving, QueuePool keeps track of the queue ("a last used queue") which prevents from
starving messages that were scheduled for the same time (basically a simple round-robin). QueuePool looks for a message
that is not scheduled to the future and can be executed (to do so we get O(N) where N is number of queues being used).

All workers listen on all queues for now. This prevents from waiting on a message that would be never processed.

In order to understand how Simulator works, you need to understand how Celery works. Please refer to Celery
documentation if you are a Celery-newbie.
"""

from datetime import datetime
from datetime import timedelta
import logging
import traceback

from celery import Task as CeleryTask
from flexmock import flexmock
from selinon import Config
from selinon import run_flow
from selinon import run_flow_selective
from selinon.systemState import SystemState
from selinonlib import UnknownError
from selinonlib.globalConfig import GlobalConfig

from .celeryMocks import simulate_apply_async
from .celeryMocks import simulate_retry
from .celeryMocks import SimulateAsyncResult
from .celeryMocks import SimulateRetry
from .progress import Progress
from .queuePool import QueuePool


class Simulator(object):
    """Simulator that simulates Selinon run in a multi-process environment."""

    simulator_queues = QueuePool()
    _logger = logging.getLogger(__name__)

    def __init__(self, nodes_definition, flow_definitions, **opts):
        """Instantiate simulator.

        :param nodes_definition: path to nodes.yaml file
        :param flow_definitions: a list of YAML files describing flows
        :param opts: additional simulator options, supported: concurrency, config_py path, sleep_time, keep_config_py
        """
        Config.set_config_yaml(nodes_definition, flow_definitions,
                               config_py=opts.pop('config_py', None),
                               keep_config_py=opts.pop('keep_config_py', False))

        self.concurrency = opts.pop('concurrency', 1)
        self.sleep_time = opts.pop('sleep_time', 1)
        self.selective = opts.pop('selective', None)

        if opts:
            raise UnknownError("Unknown options supplied: %s" % opts)

    def run(self, flow_name, node_args=None):
        """Run simulator.

        :param flow_name: a flow name that should be run
        :param node_args: arguments for the flow
        """
        # We need to assign a custom async result as we are not running Celery but our mocks instead
        flexmock(SystemState, _get_async_result=SimulateAsyncResult)
        # Overwrite used Celery functions so we do not rely on Celery logic at all
        CeleryTask.apply_async = simulate_apply_async
        CeleryTask.retry = simulate_retry

        # Let's schedule the flow - our first starting task - task will be placed onto queue - see simulate_apply_async
        if self.selective:
            run_flow_selective(
                flow_name,
                self.selective['task_names'],
                node_args,
                self.selective['follow_subflows'],
                self.selective['run_subsequent']
            )
        else:
            run_flow(flow_name, node_args)

        while not self.simulator_queues.is_empty():
            # TODO: concurrency
            self._logger.debug("new simulator run")

            # Retrieve a task that can be run right now
            time, record = self.simulator_queues.pop()
            task, celery_kwargs = record

            # we got a task with the lowest wait time - we need to wait if the task was scheduled in the future
            wait_time = (time - datetime.now()).total_seconds()
            Progress.sleep(wait_time=wait_time,
                           sleep_time=self.sleep_time,
                           info_text='Waiting for next task to process (%s seconds)... ' % round(wait_time, 3),
                           show_progressbar=self.concurrency == 1)
            try:
                kwargs = celery_kwargs.get('kwargs')
                # remove additional metadata placed by Selinon when doing tracing
                kwargs.pop('meta', None)
                result = task.run(**kwargs)

                # Dispatcher needs info about flow (JSON), but SelinonTaskEnvelope always returns None - we
                # need to keep track of success)
                SimulateAsyncResult.set_successful(task.request.id, result)
            except SimulateRetry as selinon_exc:
                if 'exc' in selinon_exc.celery_kwargs and selinon_exc.celery_kwargs.get('max_retries', 1) == 0:
                    # log only user exception as we do not want SimulateRetry in our exception traceback
                    user_exc = selinon_exc.celery_kwargs['exc']
                    user_exc_info = (user_exc, user_exc, user_exc.__traceback__)
                    self._logger.exception(str(user_exc), exc_info=user_exc_info)
                    SimulateAsyncResult.set_failed(task.request.id, traceback.format_exception(*user_exc_info))
                else:
                    # reschedule if there was an exception and we did not hit max_retries when doing retry
                    Simulator.schedule(task, selinon_exc.celery_kwargs)
            except KeyboardInterrupt:
                raise
            except Exception as exc:
                raise UnknownError("Ooooops! Congratulations! It looks like you've found a bug! Feel free to open an "
                                   "issue at https://github.com/selinon/selinonlib/issues") from exc

    @classmethod
    def schedule(cls, task, celery_kwargs):
        """Schedule a new task to be executed.

        :param task: task to be executed
        :type task: Dispatcher|SelinonTaskEnvelope
        :param celery_kwargs: arguments for the task - raw Celery arguments which also carry additional Selinon
                              arguments
        """
        cls._logger.debug("simulator is scheduling %s - %s", task.__class__.__name__, celery_kwargs)
        cls.simulator_queues.push(queue_name=celery_kwargs.get('queue', GlobalConfig.DEFAULT_CELERY_QUEUE),
                                  time=datetime.now() + timedelta(seconds=celery_kwargs.get('countdown') or 0),
                                  record=(task, celery_kwargs,))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Storage configuration and abstraction from YAML config file."""

from .cacheConfig import CacheConfig
from .errors import ConfigurationError
from .helpers import check_conf_keys


class Storage(object):
    """A storage representation."""

    def __init__(self, name, import_path, configuration, cache_config, class_name=None):
        # pylint: disable=too-many-arguments
        """Instantiate storage representation based on configuration supplied in YAML config files.

        :param name: storage name
        :param import_path: storage import path
        :param configuration: storage configuration that will be passed
        :param cache_config: cache configuration information
        :param class_name: storage class name
        """
        self.name = name
        self.import_path = import_path
        self.configuration = configuration
        self.class_name = class_name or name
        self.tasks = []
        self.cache_config = cache_config

    def register_task(self, task):
        """Register a new that uses this storage.

        :param task: task to be registered
        """
        self.tasks.append(task)

    @staticmethod
    def from_dict(dict_):
        """Construct storage instance from a dict.

        :param dict_: dict that should be used to instantiate Storage
        :rtype: Storage
        """
        if 'name' not in dict_ or not dict_['name']:
            raise ConfigurationError('Storage name definition is mandatory')
        if 'import' not in dict_ or not dict_['import']:
            raise ConfigurationError("Storage import definition is mandatory, storage '%s'" % dict_['name'])
        if 'configuration' not in dict_ or not dict_['configuration']:
            raise ConfigurationError("Storage configuration definition is mandatory, storage '%s'" % dict_['name'])
        if 'classname' in dict_ and not isinstance(dict_['classname'], str):
            raise ConfigurationError("Storage classname definition should be string, got '%s' instead, storage '%s'"
                                     % (dict_['classname'], dict_['name']))
        if 'cache' in dict_:
            if not isinstance(dict_['cache'], dict):
                raise ConfigurationError("Storage cache for storage '%s' should be a dict with configuration, "
                                         "got '%s' instead" % (dict_['name'], dict_['cache']))
            cache_config = CacheConfig.from_dict(dict_['cache'], dict_['name'])
        else:
            cache_config = CacheConfig.get_default(dict_['name'])

        # check supplied configuration options
        unknown_conf = check_conf_keys(dict_, known_conf_opts=('name', 'import', 'configuration', 'cache', 'classname'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration options for storage '%s' supplied: %s"
                                     % (dict_['name'], unknown_conf.keys()))

        return Storage(dict_['name'], dict_['import'], dict_['configuration'], cache_config, dict_.get('classname'))

    @property
    def var_name(self):
        """Return variable name which should be used for this storage reference in the generated Python code.

        return: name of storage variable that will be used in config.py file
        """
        return "_storage_%s" % self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
r"""Failure node handling representation.

A failure is basically a node in graph, that represents all permutations of possible cases of failures. Consider
having two failure conditions:

.. code-block:: yaml

  failures:
    - nodes:
          - Task1
          - Task2
          - Task3
      fallback:
          - FallbackTask1
    - nodes:
          - Task1
          - Task2
      fallback:
          - FallbackTask2

What we do here, we construct a graph of all possible permutations connected using edges that represent a node
that should be added to create a new permutation:

.. code-block:: yaml

   |-------------------------------
   |                              |
  T1           T2           T3    |
    \         /  \         / \    |
     \       /    \       /   \   |
      \     /      \     /     \  v
       T1,T2*       T2,T3      T1,T3
          \         /           /
           \       /           /
            T1,T2,T3*  <-------

For nodes ``T1,T2,T3`` and ``T1,T2`` we assign a fallback as configured. This graph is then serialized into the Python
configuration code. This way the dispatcher has O(N) time complexity when dealing with failures.

Note we are creating sparse tree - only for nodes that are listed in failures.

Note that we link failure nodes as allocated - we get a one way linked list of all failure nodes that helps us with
Python code generation.
"""

from functools import reduce

from .errors import ConfigurationError
from .predicate import Predicate


class FailureNode(object):
    """A representation of a failure node permutation."""

    def __init__(self, flow, traversed, failure_link):
        """Instantiate a failure node in the graph of all possible failures (permutations).

        :param flow: flow to which the failure node conforms to
        :param traversed: traversed nodes - also permutation of nodes
        :param failure_link: link to next failure node in failures
        """
        self.next = {}
        self.flow = flow
        self.traversed = traversed
        self.failure_link = failure_link
        self.fallbacks = []
        self.propagate_failures = []
        self.predicates = []

    def to(self, node_name):  # pylint: disable=invalid-name
        """Retrieve next permutation based on link in failure graph.

        :param node_name: a name of the node for next permutation
        :rtype: FailureNode
        """
        return self.next[node_name]

    def has_to(self, node_name):
        """Check whether there is a link to next permutation.

        :param node_name: name of the node to be added to create a new permutation
        :return: True if there is a link to next permutation for node of name node_name
        """
        return node_name in self.next

    def add_to(self, node_name, failure):
        """Add failure for next permutation.

        :param node_name: a node for next permutation
        :param failure: FailureNode that should be added
        """
        assert node_name not in self.next  # nosec
        self.next[node_name] = failure

    @staticmethod
    def _add_failure_info(failure_node, failure_info, predicate):
        """Add failure specific info to a failure node.

        :param failure_node: a failure node where the failure info should be added
        :param failure_info: additional information as passed from configuration file
        :param predicate: predicate that should be evaluated on a failure
        """
        if not isinstance(failure_info['fallback'], list) and failure_info['fallback'] is not True:
            failure_info['fallback'] = [failure_info['fallback']]

        # propagate_failure parsing
        if not isinstance(failure_info.get('propagate_failure', False), bool):
            raise ConfigurationError("Configuration option 'propagate_failure' for failure '%s' in flow '%s' "
                                     "should be boolean, got '%s' instead"
                                     % (failure_node.traversed, failure_node.flow.name,
                                        failure_info['propagate_failure']))

        if failure_info['fallback'] is True and failure_info.get('propagate_failure') is True:
            raise ConfigurationError("Configuration is misleading for failure '%s' in flow '%s' - cannot set "
                                     "propagate_failure and fallback to true at the same time"
                                     % (failure_node.traversed, failure_node.flow.name))

        failure_node.fallbacks.append(failure_info['fallback'])
        failure_node.predicates.append(predicate)
        failure_node.propagate_failures.append(failure_info.get('propagate_failure', False))

    @staticmethod
    def construct_condition_name(failure_node, idx):
        """Construct condition name that will be used in case of conditional fallbacks.

        :param failure_node: failure node for which the condition name should be constructed
        :param idx: index of condition that should be printed (could be multiple fallbacks with same source and dst)
        :return: string representation of constructed condition name as stated in the generated python code
        """
        return "_{flow}_{src}_f_{dest}_{idx}".format(flow=failure_node.flow.name,
                                                     src="_".join(failure_node.traversed),
                                                     dest="_".join(failure_node.fallbacks[idx]),
                                                     idx=idx)

    @classmethod
    def construct(cls, flow, system, failures):  # pylint: disable=too-many-locals,too-many-branches
        """Construct failures from failures dictionary.

        :param flow: flow to which failures conform to
        :param system: system context to be used
        :param failures: failures dictionary
        :return: a link for linked list of failures and a dict of starting failures
        """
        last_allocated = None
        starting_failures = {}
        predicates = []

        # pylint: disable=too-many-nested-blocks
        for failure in failures:
            used_starting_failures = {}

            for node in failure['nodes']:
                if node not in starting_failures:
                    failure_node = FailureNode(flow, [node], last_allocated)
                    last_allocated = failure_node
                    starting_failures[node] = failure_node
                    used_starting_failures[node] = failure_node
                else:
                    used_starting_failures[node] = starting_failures[node]

            current_nodes = list(used_starting_failures.values())
            next_nodes = []

            for _ in range(1, len(failure['nodes'])):  # for every permutation length

                for edge_node in failure['nodes']:  # edge_node is a node that can create a new permutation from a node

                    for current_node in current_nodes:

                        if edge_node not in current_node.traversed:

                            if not current_node.has_to(edge_node):
                                next_node = current_node.traversed + [edge_node]
                                failure_node = FailureNode(flow, next_node, last_allocated)
                                last_allocated = failure_node
                                current_node.add_to(edge_node, failure_node)

                                for node in current_nodes:
                                    diff = set(node.traversed) ^ set(next_node)

                                    if len(diff) == 1:
                                        if not node.has_to(list(diff)[0]):
                                            node.add_to(list(diff)[0], failure_node)

                                next_nodes.append(failure_node)
                            else:
                                # keep for generating new permutations
                                next_nodes.append(current_node.to(edge_node))

                current_nodes = next_nodes
                next_nodes = []

            failure_node = reduce(lambda x, y: x.to(y),
                                  failure['nodes'][1:],
                                  used_starting_failures[failure['nodes'][0]])

            if 'condition' in failure:
                nodes_from = [system.node_by_name(n) for n in failure['nodes']]
                predicate = Predicate.construct(failure['condition'], nodes_from, flow, can_inspect_results=False)
            else:
                predicate = Predicate.construct_default(flow)

            cls._add_failure_info(failure_node, failure, predicate)
            predicates.append(predicate)

        # we could make enumerable and avoid last_allocated (it would be cleaner), but let's stick with
        # this one for now
        return last_allocated, starting_failures, predicates



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""First-In-First-Out cache implementation."""

from collections import deque

from selinon import Cache
from selinon.errors import CacheMissError


class FIFO(Cache):
    """First-In-First-Out cache."""

    def __init__(self, max_cache_size):
        """Instantiate cache.

        :param max_cache_size: maximum number of items in the cache
        """
        assert max_cache_size >= 0  # nosec

        self.max_cache_size = max_cache_size
        self._cache = {}
        # Use deque as we want to do popleft() in O(1)
        self._cache_usage = deque()

    @property
    def current_cache_size(self):
        """Get current cache size.

        :return: length of the current cache
        """
        return len(self._cache_usage)

    def __repr__(self):
        """Representation of the cache for logs/debug.

        :return: a string representation
        """
        return "%s(%s)" % (self.__class__.__name__, list(self._cache_usage))

    def _clean_cache(self):
        """Trim cache."""
        while self.current_cache_size + 1 > self.max_cache_size and self.current_cache_size > 0:
            latest = self._cache_usage.popleft()
            del self._cache[latest]

    def add(self, item_id, item, task_name=None, flow_name=None):
        """Add item to cache.

        :param item_id: item id under which item should be referenced
        :param item: item itself
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        """
        if item_id in self._cache:
            return

        self._clean_cache()

        if self.max_cache_size > 0:
            self._cache[item_id] = item
            self._cache_usage.append(item_id)

    def get(self, item_id, task_name=None, flow_name=None):
        """Get item from cache.

        :param item_id: item id under which the item is stored
        :param task_name: name of task that result should/shouldn't be cached, unused when caching Celery's AsyncResult
        :param flow_name: name of flow in which task was executed, unused when caching Celery's AsyncResult
        :return: item itself
        """
        if item_id not in self._cache:
            raise CacheMissError()

        return self._cache[item_id]



#!/usr/bin/python3

import sys
import os
from setuptools import setup, find_packages


NAME = 'selinonlib'


def get_requirements():
    with open('requirements.txt') as fd:
        return fd.read().splitlines()


def get_version():
    with open(os.path.join(NAME, 'version.py')) as f:
        version = f.readline()
    # dirty, remove trailing and leading chars
    return version.split(' = ')[1][1:-2]


if sys.version_info[0] != 3:
    sys.exit("Python3 is required in order to install selinonlib")

setup(
    name=NAME,
    version=get_version(),
    packages=find_packages(),
    scripts=['selinonlib-cli'],
    install_requires=get_requirements(),
    author='Fridolin Pokorny',
    author_email='fpokorny@redhat.com',
    maintainer='Fridolin Pokorny',
    maintainer_email='fpokorny@redhat.com',
    description='a simple tool to visualize, check and generate Python code from a YAML configuration file for Selinon'
                ' dispatcher for Celery',
    url='https://github.com/selinon/selinonlib',
    license='BSD',
    keywords='node task graph edge celery selinon yaml condition',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: System :: Distributed Computing",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ]
)



#!/bin/env python3

from functools import reduce


def argsFieldExist(node_args, key):
    try:
        reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], node_args)
        return True
    except:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Configuration for caching."""

from .errors import ConfigurationError
from .helpers import check_conf_keys


class CacheConfig(object):
    """Configuration for Caching."""

    _DEFAULT_CACHE_NAME = 'LRU'
    _DEFAULT_CACHE_IMPORT = 'selinonlib.caches'
    _DEFAULT_CACHE_OPTIONS = {'max_cache_size': 0}

    def __init__(self, name, import_path, configuration, entity_name):
        """Initialize cache config as described in the YAML configuration file.

        :param name: name of the cache
        :param import_path: import from where cache should be imported
        :param configuration: cache configuration
        :param entity_name: entity for which cache should be provided
        """
        self.name = name
        self.import_path = import_path
        self.configuration = configuration
        self.entity_name = entity_name

    @property
    def var_name(self):
        """Name of variable in the generated Python code representing this cache.

        :return: name of cache variable that will be used in config.py file
        """
        return "_cache_%s_%s" % (self.entity_name, self.name)

    @classmethod
    def get_default(cls, entity_name):
        """Get default cache configuration.

        :param entity_name: entity name that will use the default cache - entity is either storage name or
                            flow name (when caching async results)
        :return: CacheConfig for the given entity
        :rtype: CacheConfig
        """
        return CacheConfig(cls._DEFAULT_CACHE_NAME, cls._DEFAULT_CACHE_IMPORT, cls._DEFAULT_CACHE_OPTIONS, entity_name)

    @classmethod
    def from_dict(cls, dict_, entity_name):
        """Parse cache configuration from a dict.

        :param dict_: dict from which cache configuration should be parsed
        :param entity_name: entity name that will use the default cache - entity is either storage name or
                            flow name (when caching async results)
        :return: cache for the given entity based on configuration
        :rtype: CacheConfig
        """
        name = dict_.get('name', cls._DEFAULT_CACHE_NAME)
        import_path = dict_.get('import', cls._DEFAULT_CACHE_IMPORT)
        configuration = dict_.get('configuration', cls._DEFAULT_CACHE_OPTIONS)

        if not isinstance(name, str):
            raise ConfigurationError("Cache configuration for '%s' expects name to be a string, got '%s' instead"
                                     % (entity_name, name))

        if not isinstance(import_path, str):
            raise ConfigurationError("Cache configuration for '%s' expects import to be a string, got '%s' instead"
                                     % (entity_name, import_path))

        if not isinstance(configuration, dict):
            raise ConfigurationError("Cache configuration for '%s' expects configuration to be a dict of cache "
                                     "configuration, got '%s' instead" % (entity_name, configuration))

        # check supplied configuration configuration
        unknown_conf = check_conf_keys(dict_, known_conf_opts=('name', 'import', 'configuration'))
        if unknown_conf:
            raise ConfigurationError("Unknown configuration configuration for cache '%s' supplied: %s"
                                     % (name, unknown_conf))

        return CacheConfig(name, import_path, configuration, entity_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Injected Celery related implementations of methods.

Classes and functions to make Selinon simulator work as a standalone CLI.
"""

from flexmock import flexmock


class SimulateRequest(object):
    """Simulate Celery's Task.request.

    Make possible to query task id right inside task by calling self.request.id
    """

    def __init__(self, instance):
        """Instantiate request.

        :param instance:Instance for the request.
        """
        self.id = str(id(instance))  # pylint: disable=redefined-builtin,invalid-name


class SimulateAsyncResult(object):
    """Simulate AsyncResult returned by apply_async() or by instantiating AsyncResult by Task id."""

    task_failures = {}
    task_successes = {}

    def __init__(self, node_name, node_id):  # pylint: disable=redefined-builtin,invalid-name
        """Initialize AsyncResult.

        :param node_name: name of the node for which the AsyncResult should be stored
        :param node_id: ID of node for which the AsyncResult should be stored
        """
        self.task_id = str(node_id)
        self.node_id = node_name

    @classmethod
    def set_successful(cls, task_id, result):
        """Mark task as successful after run - called by simulator.

        :param task_id: an ID of task that should be marked as successful
        :param result: result of task (None for SelinonTaskEnvelope, JSON describing system state for Dispatcher)
        """
        cls.task_successes[task_id] = result

    @classmethod
    def set_failed(cls, task_id, exception):
        """Mark task as failed after run - called by simulator.

        :param task_id: an ID of task that should be marked as failed
        :param exception: exception raised in task
        """
        cls.task_failures[task_id] = exception

    def successful(self):
        """Check for success.

        :return: True if task succeeded.
        """
        return self.task_id in self.task_successes

    def failed(self):
        """Check for failure.

        :return: True if task failed
        """
        return self.task_id in self.task_failures

    @property
    def traceback(self):
        """Traceback as returned by Celery's AsyncResult.

        :return: traceback returned by a task
        """
        return self.task_failures[self.task_id]

    @property
    def result(self):
        """Get result of a task.

        :return: retrieve result of the task or exception that was raised
        """
        return self.task_failures.get(self.task_id, None)


class SimulateRetry(Exception):
    """Simulate Celery Retry exception raised by self.retry()."""

    def __init__(self, instance, **celery_kwargs):
        """Instantiate Retry exception.

        :param instance: instance that triggered retry
        :param celery_kwargs: kwargs arguments as passed to raw Celery task
        """
        super().__init__()
        self.instance = instance
        self.celery_kwargs = celery_kwargs


def simulate_apply_async(instance, **celery_kwargs):
    """Simulate CeleryTask().apply_async() implementation for scheduling tasks.

    :param instance: instance that should be scheduled
    :param celery_kwargs: kwargs supplied to Celery Task (also carry arguments for Selinon)
    """
    from .simulator import Simulator

    flexmock(instance, request=SimulateRequest(instance))
    Simulator.schedule(instance, celery_kwargs)
    selinon_kwargs = celery_kwargs['kwargs']
    return SimulateAsyncResult(selinon_kwargs.get('task_name', selinon_kwargs['flow_name']),
                               node_id=id(instance))


def simulate_retry(instance, **celery_kwargs):
    """Simulate Celery self.retry() implementation for retrying tasks.

    :param instance: instance that should called self.retry()
    :param celery_kwargs: kwargs that will be supplied to Celery Task (also carry arguments for Selinon)
    """
    raise SimulateRetry(instance, **celery_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Supporting routines for run time."""


def always_run(flow_name, node_name, node_args, task_names, storage_pool):  # pylint: disable=unused-argument
    """Run the default function that is called on selective run.

    :param flow_name: flow name in which the selective run is done
    :param node_name: name of the node on which this function was run
    :param node_args: arguments supplied to the flow
    :param task_names: a list of nodes that should be run in the selective run
    :param storage_pool: storage pool for the parent nodes
    :return: None if desired node should be run, id of task that results should be reused
    """
    return None



#!/usr/bin/env python3

from functools import reduce
from urllib.parse import urlparse


def argsFieldUrlNetloc(node_args, key, netloc):
    try:
        val = reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], node_args)
        return urlparse(val).netloc == netloc
    except:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
from selinon.errors import CacheMissError
from selinonlib.caches import FIFO
from selinonlibTestCase import SelinonlibTestCase


class TestFIFO(SelinonlibTestCase):
    """
    Test First-In-First-Out cache
    """
    @staticmethod
    def _item_id2item(i):
        return "x%s" % i

    def test_one_item_miss(self):
        cache = FIFO(max_cache_size=1)

        cache.add("item_id1", "item1", "Task1", "flow1")
        cache.add("item_id2", "item2", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            cache.get("item_id1", "Task1", "flow1")

    def test_two_items(self):
        cache = FIFO(max_cache_size=2)

        cache.add("item_id1", "item1", "Task1", "flow1")
        cache.add("item_id2", "item2", "Task1", "flow1")
        cache.add("item_id3", "item3", "Task1", "flow1")

        with pytest.raises(CacheMissError):
            cache.get("item_id1", "Task1", "flow1")

    def test_multiple_items(self):
        item_count = 16
        cache = FIFO(max_cache_size=item_count)

        for item_id in range(item_count):
            cache.add(item_id, self._item_id2item(item_id), "Task1", "flow1")

        cache.add(item_count, self._item_id2item(item_count), "Task1", "flow1")

        with pytest.raises(CacheMissError):
            # first out
            cache.get(0, "Task1", "flow2")

    def test_multiple_items2(self):
        item_count = 16
        cache = FIFO(max_cache_size=item_count)

        for item_id in range(item_count):
            cache.add(item_id, self._item_id2item(item_id), "Task1", "flow1")

        for item_id in range(item_count - 1, -1, -1):
            assert cache.get(item_id, "Task1", "flow1") == self._item_id2item(item_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2017  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Indicate progress and sleep for given time."""

from math import ceil
import sys
from time import sleep


class Progress(object):
    """Indicate progress and sleep for given time."""

    _indicators = ('-', '\\', '|', '/')
    _current_indicator_idx = 0

    @classmethod
    def indicate(cls, iterable, show_progressbar=True, info_text=None):
        """Indicate progress on iterable.

        :param iterable: iterable that is used to iterate on progress
        :param show_progressbar: if True, there is shown a simple ASCII art spinning
        :param info_text: text that is printed on the line (progressbar follows)
        """
        for item in iterable:
            sys.stdout.write(info_text or '')
            if show_progressbar:
                sys.stdout.write(cls._indicators[cls._current_indicator_idx])
                cls._current_indicator_idx = (cls._current_indicator_idx + 1) % len(cls._indicators)

            sys.stdout.flush()
            yield item
            sys.stdout.write('\r')

        # clear whole console line at the end
        sys.stdout.write("\033[K")

    @classmethod
    def sleep(cls, wait_time, sleep_time, info_text=None, show_progressbar=True):
        """Wait and sleep for the given amount of time.

        :param wait_time: time to wait in this method in total
        :param sleep_time: time between periodic checks (parameter to sleep() function)
        :param show_progressbar: if True, there is shown a simple ASCII art spinning
        :param info_text: text that is printed on the line (progressbar follows)
        """
        total_wait_time = ceil(wait_time / sleep_time)
        for _ in cls.indicate(range(total_wait_time), show_progressbar, info_text=info_text,):
            sleep(sleep_time)
